from netgen.meshing import *
from netgen.csg import *
from ngsolve import *
from math import pi
import scipy.sparse.linalg
import numpy as np
from scipy.sparse import csr_matrix
import sys,os,copy 
sys.path.append(os.path.realpath(''))
from sweeping_geom import Layered_Mesh,Make1DMesh,Make1DMesh_refined  
from sweeping_helper import load_PREM,BitArrayFromList,P_DoFs

ngsglobals.msg_level = 0

#########################################################################
# Parameters (to be changed by the user) 
#########################################################################

use_PML = False # set 'True' to set a PML at Earth's surface
                # and 'False' for the free surface boundary condition

#########################################################################
# Further parameters
#########################################################################

eps_perturb = 3.90625*10**-5 # strength of perturbation 
order =  4   # order of FEM 
order_ODE = order # order of ODE discretization
bonus_intorder = 0
C_PML = 100 # amplitude of PML profile
solver = "umfpack" # direct solver for ODEs
omega2layers = {256:3,
                512:6,
                1024:12,
                2048:24} # mapping of wavenumber to layers (J)

#########################################################################
# Coefficients of problem 
#########################################################################

# load PREM coefficients 
c_ref,c_1D,rho,rho_1D = load_PREM()
c = c_ref
c_1D_ref = c_1D 
c_perturb_1D = 1.0
c_eps_1D = c_1D_ref*(1+eps_perturb*c_perturb_1D)  

# weights
weight_r = y**2
weight_r_1D = x**2
weight_theta = sin(x)**3
weight_factor = CoefficientFunction(weight_r * weight_theta)

#########################################################################
# Computation of DtN numbers (ODE solves) 
#########################################################################

def precompute_DtN_coeff(omega,c_1D):

    #########################################################################
    # Layered mesh
    #########################################################################

    R_earth = 6371
    r_CMB = (R_earth - 2891) / R_earth
    n_layers = omega2layers[omega]
    r_layer =  [ r_CMB + (1.0-r_CMB)*i/(n_layers) for i in range(n_layers+1)] 
    #nx = int(2*n_layers) # this would be used in practice
    nx = int(2*24) # for demonstration we use the same mesh for all wavenumbers
    mesh = Layered_Mesh(r_layer,nx,True,[False for i in range(len(r_layer))],'Dirichlet')
    
    #########################################################################
    # FEM space (Proxy and-GridFunctions) 
    #########################################################################

    fes_H1 = H1(mesh, complex=True,  order=order,dirichlet="left|right")
    fes = fes_H1

    free_dofs = [] 
    for nr,b in enumerate(fes.FreeDofs()): 
        if b:
            free_dofs.append(nr)

    #########################################################################
    # Domain decomposition 
    #########################################################################
    
    rr = r_layer
    r_layer = rr[1:]
    intervals = [ (rr[i],rr[i+1]) for i in range(len(rr)-1)]
    intervals = intervals[::-1]

    # compute centroids of elements    
    centroids = []
    for el in mesh.Elements():
        vs = el.vertices
        L = len(el.vertices)
        c_x = 0
        c_y = 0
        for vert in el.vertices:
            vx,vy,vz = mesh.ngmesh.Points()[vert.nr+1].p
            c_x = c_x + vx
            c_y = c_y + vy
        c_x = c_x / L
        c_y = c_y / L
        centroids.append((c_x,c_y))

    layer2el = [ [] for i in range(n_layers)]         #elements of layer
    layer2dof = [ [] for i in range(n_layers)]        #dofs of layer   
    IF2dof = [ [] for i in range(n_layers)]           #dofs of layer 
    Interior2dof = [ [] for i in range(n_layers)] 
    InteriorPlusInnerdof = [ [] for i in range(n_layers)] 
    layer2all_outer_dof = [ [] for i in range(n_layers)] 
    layer2all_outer_el  = [ [] for i in range(n_layers)] 
    IF2el = [ [] for i in range(n_layers)] # elements on interface


    def centroid2layer(centr):
        cx,cy = centr
        r_c =  cy
        for l,I in enumerate(intervals):
            if r_c >= I[0] and r_c <= I[1]:
                return l
      
    for cent,el in zip(centroids,fes.Elements()):
        mylayer = centroid2layer(cent)
        layer2el[mylayer].append(el.nr)
        for dof in el.dofs:
            layer2dof[mylayer].append(dof)
    for i in range(n_layers):
        layer2dof[i] = list( set(layer2dof[i]).intersection(free_dofs)  )
        layer2dof[i].sort()
    for i in range(1,n_layers):
        IF2dof[i] = list( set(layer2dof[i]).intersection( set(layer2dof[i-1])) )
        IF2dof[i].sort()
    for i in range(1,n_layers-1):
         Interior2dof[i] = list( set(layer2dof[i]) - set(IF2dof[i]) - set(IF2dof[i+1])  )
         Interior2dof[i].sort()
    Interior2dof[n_layers-1] = list( set(layer2dof[n_layers-1]) - set(IF2dof[n_layers-1])  )
    Interior2dof[n_layers-1].sort()
    Interior2dof[0] = list( set(layer2dof[0]) - set(IF2dof[1])  )
    for i in range(1,n_layers):
        for j in range(0,i-1):
            layer2all_outer_dof[i].extend(layer2dof[j])
        layer2all_outer_dof[i].extend(Interior2dof[i-1])
        layer2all_outer_dof[i] =  list( set(layer2all_outer_dof[i]))
        layer2all_outer_dof[i].sort()
    for i in range(1,n_layers):
        for j in range(0,i):
            layer2all_outer_el[i].extend(layer2el[j])
        layer2all_outer_el[i] =  list( set(layer2all_outer_el[i]))
    for i in range(n_layers):
        InteriorPlusInnerdof[i] = list( set(Interior2dof[i]) | set(IF2dof[i])  )
        InteriorPlusInnerdof[i].sort()

    u,v = fes.TnT()
    
    #########################################################################
    # PML 
    #########################################################################
    
    f0 = omega / 2 
    rad_min = r_layer[-2]
    rad_max = r_layer[-1]
    eta = rad_max-rad_min
    
    physpml_sigma_1D = IfPos(x-rad_min,IfPos(rad_max -x,C_PML/eta*(-rad_max+x+eta)**2/eta**2,0),0)
    pml_phys_1D = pml.Custom(trafo=x+(1j/f0)*(C_PML/3)*(eta/C_PML)**(3/2)*physpml_sigma_1D**(3/2), jac = -1 - 1j*physpml_sigma_1D/f0 )
    C_phys = C_PML
    
    #########################################################################
    # Computation of DtN numbers (ODE solves) for fixed omega
    #########################################################################
    
    DtN_coeff_layer = [ [] for i in range(n_layers)]  
    DtN_PML_coeff_layer = [ [] for i in range(n_layers)]  
    lam_layer = [ None for i in range(n_layers)] 
    lam_vec_layer = [ None for i in range(n_layers)] 

    R_max = r_layer[-1]

    for layer_nr in range(1,n_layers):

        if_idx = layer_nr
        ba_dofs = BitArrayFromList(IF2dof[if_idx],mesh,fes,elements=False)
        m_IF = BilinearForm (fes, symmetric=False, check_unused=False)
        m_IF += SymbolicBFI ( u*v*weight_theta, definedon=mesh.Boundaries("layer"+str(if_idx)+"-bnd"),bonus_intorder =  bonus_intorder )
        m_IF.Assemble()
        k_IF = BilinearForm (fes, symmetric=False, check_unused=False)
        k_IF += SymbolicBFI ( grad(u).Trace()*grad(v).Trace()*weight_theta, definedon=mesh.Boundaries("layer"+str(if_idx)+"-bnd"), bonus_intorder =  bonus_intorder) 
        k_IF.Assemble()
        
        rows_m,cols_m,vals_m = m_IF.mat.COO()
        M_if = P_DoFs(csr_matrix((vals_m,(rows_m,cols_m))),IF2dof[if_idx],IF2dof[if_idx]).todense() 
        rows_k,cols_k,vals_k = k_IF.mat.COO()
        K_if = P_DoFs(csr_matrix((vals_k,(rows_k,cols_k))),IF2dof[if_idx],IF2dof[if_idx]).todense()

        lam,lam_vec =  scipy.linalg.eig(K_if, b=M_if, left=False, right=True, overwrite_a=False, overwrite_b=False, check_finite=True)
       
        sort_idx = np.argsort(np.abs(lam))
        lam = lam[sort_idx]
        lam_vec = lam_vec[sort_idx]
    
        lam_diag = np.diag(lam)

        lam_vec = lam_vec @ np.diag(  1/np.sqrt(np.diag( lam_vec.conj().T @ (M_if @ lam_vec))) )
        lam_layer[layer_nr] = lam.copy()
        lam_vec_layer[layer_nr] = lam_vec.copy()

        mesh_1D = Make1DMesh(layer_nr,r_layer)
        fes_DtN = H1(mesh_1D, complex=True,  order=order_ODE, dirichlet=[1])
        u_DtN,v_DtN = fes_DtN.TnT()
        gfu_DtN = GridFunction (fes_DtN)
        f_DtN = LinearForm(fes_DtN)
        f_DtN.Assemble()
        r_DtN = f_DtN.vec.CreateVector()
        frees_DtN = [i for i in range(1,fes_DtN.ndof)]
        DtN_exact_plot = []

        for lami in lam:
            if use_PML:
                mesh_1D.SetPML(pml_phys_1D,definedon=1)

            a_DtN = BilinearForm (fes_DtN, symmetric=False, check_unused=False)
            a_DtN += SymbolicBFI( x**2*(rho_1D*c_1D**2)*grad(u_DtN)*grad(v_DtN)*weight_r_1D, bonus_intorder=  bonus_intorder)  
            a_DtN += SymbolicBFI(  ( (rho_1D*c_1D**2)*lami.item() - (rho_1D*omega**2*x**2) )*weight_r_1D*u_DtN*v_DtN, bonus_intorder= bonus_intorder)
            
            # for constant coefficient problem
            #a_DtN += SymbolicBFI(  grad(u_DtN)*grad(v_DtN), bonus_intorder=  bonus_intorder)  
            #a_DtN += SymbolicBFI(   (CoefficientFunction(1.0)*lami.item() - c_1D**2*omega**2 )*u_DtN*v_DtN, bonus_intorder= bonus_intorder)
            
            a_DtN.Assemble()

            if use_PML:
                mesh_1D.UnSetPML(1)
            
            gfu_DtN.vec[:] = 0.0
            gfu_DtN.vec[0] = 1.0
            r_DtN.data = f_DtN.vec - a_DtN.mat * gfu_DtN.vec
            gfu_DtN.vec.data += a_DtN.mat.Inverse(freedofs=fes_DtN.FreeDofs()) * r_DtN
            
            rows_DtN,cols_DtN,vals_DtN = a_DtN.mat.COO()
            A_DtN = scipy.sparse.csr_matrix((vals_DtN,(rows_DtN,cols_DtN)))
            val1 = P_DoFs(A_DtN,[0],frees_DtN).dot(gfu_DtN.vec.FV().NumPy()[frees_DtN])
            val2 = P_DoFs(A_DtN,[0],[0]).dot(gfu_DtN.vec.FV().NumPy()[0])
            val = -P_DoFs(A_DtN,[0],frees_DtN).dot(gfu_DtN.vec.FV().NumPy()[frees_DtN])-P_DoFs(A_DtN,[0],[0]).dot(gfu_DtN.vec.FV().NumPy()[0])

            DtN_coeff_layer[layer_nr].append(-val.item())
            #DtN_coeff_layer[layer_nr].append(-val.item()/R_min**2)

    return DtN_coeff_layer,lam_layer,lam_vec_layer


#########################################################################
# Loop over different wavenumbers 
#########################################################################

for omega in [512,1024,2048]:
    # DtN numbers of background model
    DtN_coeff_layer,lam_layer,lam_vec_layer = precompute_DtN_coeff(omega,c_1D_ref)
    # DtN numbers of perturbed model
    DtN_coeff_layer_eps,lam_layer_eps,lam_vec_layer_eps = precompute_DtN_coeff(omega,c_eps_1D)

    # Consider interface closest to the CMB -> [-1]
    DtN_nr_ref = np.array(DtN_coeff_layer[-1])
    DtN_nr_eps = np.array(DtN_coeff_layer_eps[-1])
    print("(mode,relative error) for omega =",omega)   
    err_modes = (np.abs(DtN_nr_ref-DtN_nr_eps)/np.abs(DtN_nr_ref)).tolist()
    for lam_idx,err in enumerate(err_modes):
        print("({0},{1})".format(lam_idx,err))
    #print("relative error  = ",err_modes )
    input("Press any key to continue")
