from netgen.meshing import *
from netgen.csg import *
from ngsolve import *
#import ngsolve
from math import pi
import scipy.sparse.linalg
import numpy as np
from ngsolve.solvers import GMRes
from decimal import Decimal
from scipy.linalg import lu_factor,lu_solve
from scipy.sparse import csr_matrix
import sys,os,copy 
sys.path.append(os.path.realpath(''))
from sweeping_geom import Layered_Mesh,Make1DMesh,Make1DMesh_refined  
from sweeping_helper import load_PREM,BitArrayFromList,P_DoFs

ngsglobals.msg_level = 0

#########################################################################
# Parameters (to be changed by the user)
#########################################################################

omega = 1024 # wavenumber
n_layers = 12 # number of layers (J in the paper)
eps_perturbs = [0.0,0.000625,0.00125,0.0025,0.005,0.01,0.02,0.04,0.08] 
eps_perturb = eps_perturbs[0] # strength of sound speed perturbation
sources = ["Random","Dirac"] 
source = sources[0] # right hand side: Random or Dirac source
zero_source_layer0 = False  # Set 'True' to set the random source to zero on the 
                           # first layer (if 'use_PML == True' this will be 
                           # done automatically)

use_PML = False # set 'True' to set a PML at Earth's surface
               # and False for the free surface boundary condition
use_moving_PML = False  # set 'True' to use moving PML approximation of the DtN
                        # and 'False' to use the tensor product DtN
create_vtk_output = False # set 'True' to create vtk output of direct solver sol.

#########################################################################
# Further parameters
#########################################################################

order =  4   # order of FEM 
order_ODE = order # order of ODE discretization
bonus_intorder = 0 
nx = int(2*n_layers) # number of elements in horizontal direction
solver = "umfpack" # direct solver (e.g. used for subdomain problems)
tol = 1e-7 # tolerance for GMRES
C_PML = 100 # amplitude of PML profile

#########################################################################
# Layered mesh
#########################################################################

R_earth = 6371 # radius of Earth in km
r_CMB = (R_earth - 2891) / R_earth # location of CMB
r_layer =  [ r_CMB + (1.0-r_CMB)*i/(n_layers) for i in range(n_layers+1)] # vertices of layers
mesh = Layered_Mesh(r_layer,nx,True,[False for i in range(len(r_layer))],'Dirichlet')

#########################################################################
# Coefficients of problem 
#########################################################################

# load PREM model
c_ref,c_1D,rho,rho_1D = load_PREM()
c_perturb = cos(x*y)*sin(x*y) # perturbation
c = c_ref*(1+eps_perturb*c_perturb) 

# weights
weight_r = y**2
weight_r_1D = x**2
weight_theta = sin(x)**3
weight_factor = CoefficientFunction(weight_r * weight_theta)

#########################################################################
# FEM space (Proxy and-GridFunctions) 
#########################################################################

fes = H1(mesh, complex=True,  order=order,dirichlet="left|right")
free_dofs = [] 
for nr,b in enumerate(fes.FreeDofs()): 
    if b:
        free_dofs.append(nr)

u,v = fes.TnT()
gfu = GridFunction (fes)
gfu_GMRES = GridFunction (fes)
udirect = GridFunction (fes)
gfu_DOSM_in = GridFunction (fes)
gfu_DOSM_out = GridFunction (fes)
gfu_DOSM_interm = GridFunction (fes)

#########################################################################
# Domain decomposition 
#########################################################################

rr = r_layer
r_layer = rr[1:]
intervals = [ (rr[i],rr[i+1]) for i in range(len(rr)-1)]
intervals = intervals[::-1]

# compute centroids of elements    
centroids = []
for el in mesh.Elements():
    vs = el.vertices
    L = len(el.vertices)
    c_x = 0
    c_y = 0
    for vert in el.vertices:
        vx,vy,vz = mesh.ngmesh.Points()[vert.nr+1].p
        c_x = c_x + vx
        c_y = c_y + vy
    c_x = c_x / L
    c_y = c_y / L
    centroids.append((c_x,c_y))

layer2el = [ [] for i in range(n_layers)]         #elements of layer
layer2dof = [ [] for i in range(n_layers)]        #dofs of layer   
IF2dof = [ [] for i in range(n_layers)]           #dofs of interfaces
Interior2dof = [ [] for i in range(n_layers)] 
InteriorPlusInnerdof = [ [] for i in range(n_layers)] 
layer2all_outer_dof = [ [] for i in range(n_layers)] 
layer2all_outer_el  = [ [] for i in range(n_layers)] 
IF2el = [ [] for i in range(n_layers)] # elements on interface

def centroid2layer(centr):
    cx,cy = centr
    r_c =  cy
    for l,I in enumerate(intervals):
        if r_c >= I[0] and r_c <= I[1]:
            return l

for cent,el in zip(centroids,fes.Elements()):
    mylayer = centroid2layer(cent)
    layer2el[mylayer].append(el.nr)
    for dof in el.dofs:
        layer2dof[mylayer].append(dof)
for i in range(n_layers):
    layer2dof[i] = list( set(layer2dof[i]).intersection(free_dofs)  )
    layer2dof[i].sort()
for i in range(1,n_layers):
    IF2dof[i] = list( set(layer2dof[i]).intersection( set(layer2dof[i-1])) )
    IF2dof[i].sort()
for i in range(1,n_layers-1):
     Interior2dof[i] = list( set(layer2dof[i]) - set(IF2dof[i]) - set(IF2dof[i+1])  )
     Interior2dof[i].sort()
Interior2dof[n_layers-1] = list( set(layer2dof[n_layers-1]) - set(IF2dof[n_layers-1])  )
Interior2dof[n_layers-1].sort()
Interior2dof[0] = list( set(layer2dof[0]) - set(IF2dof[1])  )
for i in range(1,n_layers):
    for j in range(0,i-1):
        layer2all_outer_dof[i].extend(layer2dof[j])
    layer2all_outer_dof[i].extend(Interior2dof[i-1])
    layer2all_outer_dof[i] =  list( set(layer2all_outer_dof[i]))
    layer2all_outer_dof[i].sort()
for i in range(1,n_layers):
    for j in range(0,i):
        layer2all_outer_el[i].extend(layer2el[j])
    layer2all_outer_el[i] =  list( set(layer2all_outer_el[i]))
for i in range(n_layers):
    InteriorPlusInnerdof[i] = list( set(Interior2dof[i]) | set(IF2dof[i])  )
    InteriorPlusInnerdof[i].sort()

#########################################################################
# PML 
#########################################################################

f0_solve = omega / 2    
f0 = f0_solve

if use_PML:

    definedon_PML = 1
    rad_min = r_layer[-2]
    rad_max = r_layer[-1]
    eta = rad_max-rad_min
    physpml_sigma = IfPos(y-rad_min,IfPos(rad_max-y,C_PML/eta*(-rad_max+y+eta)**2/eta**2,0),0) 
    physpml_trafo = CoefficientFunction( (x,
                                     y+(1j/f0)*(C_PML/3)*(eta*physpml_sigma/C_PML)**(3/2) ) )
    physpml_jac = CoefficientFunction((1.0,
                                   0.0,
                                   0.0,
                                   1+(1j/f0)*physpml_sigma 
                                   ),
                              dims=(2,2))   

    physpml_sigma_1D = IfPos(x-rad_min,IfPos(rad_max -x,C_PML/eta*(-rad_max+x+eta)**2/eta**2,0),0)
    pml_phys_1D = pml.Custom(trafo=x+(1j/f0)*(C_PML/3)*(eta/C_PML)**(3/2)*physpml_sigma_1D**(3/2), jac = -1 - 1j*physpml_sigma_1D/f0 )

    pml_phys = pml.Custom(trafo=physpml_trafo, jac = physpml_jac ) 

#########################################################################
# Solution with direct solver  
#########################################################################

if use_PML:
    mesh.SetPML(pml_phys,definedon=definedon_PML)

a = BilinearForm (fes, symmetric=False)
a += SymbolicBFI ( ( y**2*(rho*c**2)*grad(u)[1]*grad(v)[1] +  (rho*c**2)*grad(u)[0]*grad(v)[0])*weight_factor, bonus_intorder = bonus_intorder  )
a += SymbolicBFI (-rho*omega**2*y**2*u*v*weight_factor,  bonus_intorder = bonus_intorder)
a.Assemble()

if use_PML:
    mesh.UnSetPML(definedon_PML)

f = LinearForm (fes)
f.Assemble()

if source == "Random":
    print("using rdm source")
    np.random.seed(123)
    rhs_random = np.random.rand(fes.ndof)
    f.vec.FV().NumPy()[free_dofs] = 0.001*rhs_random[free_dofs]
    if use_PML or zero_source_layer0: # set source to zero in PML layer
        f.vec.FV().NumPy()[layer2dof[0]] = 0.0
else:
    for i in range(fes.ndof):
        gfu.vec[i] = 1.0
        f.vec[i] =  100*gfu(mesh(pi/2,0.7))
        gfu.vec[i] = 0.0

gfu.vec[:] = 0.0
res_direct = gfu.vec.CreateVector()
udirect.vec[:] = 0.0
res_direct.data = f.vec - a.mat * udirect.vec
invA = a.mat.Inverse(freedofs=fes.FreeDofs(), inverse=solver)
udirect.vec.data += invA * res_direct
print("Solution with direct solver")
Draw (udirect, mesh, "direct")

if create_vtk_output:
    vtk = VTKOutput(ma=mesh,coefs=[udirect.real],
                    names=["sol"],
                    filename="vtk_SH-wave",
                    subdivision=4)
    vtk.Do()

#########################################################################
# Computation of DtN numbers (ODE solves) 
#########################################################################

def precompute_DtN_coeff():

    DtN_coeff_layer = [ [] for i in range(n_layers)]  
    DtN_PML_coeff_layer = [ [] for i in range(n_layers)]  
    lam_layer = [ None for i in range(n_layers)] 
    lam_vec_layer = [ None for i in range(n_layers)] 


    R_max = r_layer[-1]

    for layer_nr in range(1,n_layers):

        if_idx = layer_nr
        ba_dofs = BitArrayFromList(IF2dof[if_idx],mesh,fes,elements=False)
        m_IF = BilinearForm (fes, symmetric=False, check_unused=False)
        m_IF += SymbolicBFI ( u*v*weight_theta, definedon=mesh.Boundaries("layer"+str(if_idx)+"-bnd"),bonus_intorder =  bonus_intorder )
        m_IF.Assemble()
        k_IF = BilinearForm (fes, symmetric=False, check_unused=False)
        k_IF += SymbolicBFI ( grad(u).Trace()*grad(v).Trace()*weight_theta, definedon=mesh.Boundaries("layer"+str(if_idx)+"-bnd"), bonus_intorder =  bonus_intorder) 
        k_IF.Assemble()
        
        rows_m,cols_m,vals_m = m_IF.mat.COO()
        M_if = P_DoFs(csr_matrix((vals_m,(rows_m,cols_m))),IF2dof[if_idx],IF2dof[if_idx]).todense() 
        rows_k,cols_k,vals_k = k_IF.mat.COO()
        K_if = P_DoFs(csr_matrix((vals_k,(rows_k,cols_k))),IF2dof[if_idx],IF2dof[if_idx]).todense()


        lam,lam_vec =  scipy.linalg.eig(K_if, b=M_if, left=False, right=True, overwrite_a=False, overwrite_b=False, check_finite=True)
        
        lam_diag = np.diag(lam)

        lam_vec = lam_vec @ np.diag(  1/np.sqrt(np.diag( lam_vec.conj().T @ (M_if @ lam_vec))) )
 
        lam_layer[layer_nr] = lam.copy()
        lam_vec_layer[layer_nr] = lam_vec.copy()

         
        R_min = r_layer[-1-layer_nr]
        mesh_1D = Make1DMesh(layer_nr,r_layer)
        fes_DtN = H1(mesh_1D, complex=True,  order=order_ODE, dirichlet=[1])
        u_DtN,v_DtN = fes_DtN.TnT()
        gfu_DtN = GridFunction (fes_DtN)
        f_DtN = LinearForm(fes_DtN)
        f_DtN.Assemble()
        r_DtN = f_DtN.vec.CreateVector()
        frees_DtN = [i for i in range(1,fes_DtN.ndof)]

        DtN_exact_plot = []

        for lami in lam:
            if use_PML:
                mesh_1D.SetPML(pml_phys_1D,definedon=1)
            a_DtN = BilinearForm (fes_DtN, symmetric=False, check_unused=False)
            a_DtN += SymbolicBFI( x**2*(rho_1D*c_1D**2)*grad(u_DtN)*grad(v_DtN)*weight_r_1D, bonus_intorder=  bonus_intorder)  
            a_DtN += SymbolicBFI(  ( (rho_1D*c_1D**2)*lami.item() - (rho_1D*omega**2*x**2) )*weight_r_1D*u_DtN*v_DtN, bonus_intorder= bonus_intorder)
            
            a_DtN.Assemble()

            if use_PML:
                mesh_1D.UnSetPML(1)
            
            gfu_DtN.vec[:] = 0.0
            gfu_DtN.Set(1.0, BND)
            r_DtN.data = f_DtN.vec - a_DtN.mat * gfu_DtN.vec
            gfu_DtN.vec.data += a_DtN.mat.Inverse(freedofs=fes_DtN.FreeDofs()) * r_DtN
            
            rows_DtN,cols_DtN,vals_DtN = a_DtN.mat.COO()
            A_DtN = scipy.sparse.csr_matrix((vals_DtN,(rows_DtN,cols_DtN)))
            val1 = P_DoFs(A_DtN,[0],frees_DtN).dot(gfu_DtN.vec.FV().NumPy()[frees_DtN])
            val2 = P_DoFs(A_DtN,[0],[0]).dot(gfu_DtN.vec.FV().NumPy()[0])
            val = -P_DoFs(A_DtN,[0],frees_DtN).dot(gfu_DtN.vec.FV().NumPy()[frees_DtN])-P_DoFs(A_DtN,[0],[0]).dot(gfu_DtN.vec.FV().NumPy()[0])

            DtN_coeff_layer[layer_nr].append(-val.item())
            #DtN_coeff_layer[layer_nr].append(-val.item()/R_min**2)


    return DtN_coeff_layer,lam_layer,lam_vec_layer

DtN_coeff_layer,lam_layer,lam_vec_layer = precompute_DtN_coeff()

#########################################################################
# Computation of layer matrices and factorizations
#########################################################################

layermat = []
layerinverse = []
right_hand_side = f.vec

for layer in range(n_layers):
    layer_name = "layer"+str(layer)
    if use_PML:
        mesh.SetPML(pml_phys,definedon=definedon_PML)
    a_layer = BilinearForm (fes, symmetric=False, check_unused=False)
    ba_els = BitArrayFromList(layer2el[layer],mesh,fes,elements=True) 
    a_layer += SymbolicBFI ( ( y**2*(rho*c**2)*grad(u)[1]*grad(v)[1] +  (rho*c**2)*grad(u)[0]*grad(v)[0])*weight_factor ,definedonelements= ba_els,bonus_intorder =  bonus_intorder)  
    a_layer += SymbolicBFI ( -rho*omega**2*y**2*u*v*weight_factor ,definedonelements= ba_els,bonus_intorder =  bonus_intorder)  
    a_layer.Assemble()
    layermat.append(a_layer.mat)
    ba_dofs = BitArrayFromList(Interior2dof[layer],mesh,fes,elements=False)
    layerinverse.append(a_layer.mat.Inverse(ba_dofs,inverse=solver))
    if use_PML:
        mesh.UnSetPML(definedon_PML)

layermat_outer = [None for i in range(n_layers)]
layerinverse_outer= [None for i in range(n_layers)]
for layer in range(1,n_layers):
    layer_name = "layer0"
    for j in range(1,layer):
        layer_name += "|layer"+str(j)
    if use_PML:
        mesh.SetPML(pml_phys,definedon=definedon_PML)
    ba_els = BitArrayFromList(layer2all_outer_el[layer],mesh,fes,elements=True) 
    
    a_outer = BilinearForm (fes, symmetric=False, check_unused=False)
    a_outer += SymbolicBFI ( ( y**2*(rho*c**2)*grad(u)[1]*grad(v)[1] +  (rho*c**2)*grad(u)[0]*grad(v)[0])*weight_factor  ,definedonelements= ba_els,bonus_intorder =  bonus_intorder) 
    a_outer += SymbolicBFI ( -rho*omega**2*y**2*u*v*weight_factor , definedonelements= ba_els,bonus_intorder =  bonus_intorder) 

    a_outer.Assemble()
    layermat_outer[layer] = a_outer.mat
    ba_dofs = BitArrayFromList(layer2all_outer_dof[layer],mesh,fes,elements=False)
    #layerinverse_outer[layer] = a_outer.mat.Inverse(ba_dofs,inverse=solver)
    if use_PML:
        mesh.UnSetPML(definedon_PML)

if use_moving_PML:
    layermat_moving_PML = [None for i in range(n_layers)]
    layerinverse_moving_PML= [None for i in range(n_layers)]
    for layer in range(1,n_layers):
        
        if layer <=1: 
            C_sweep = C_PML
        else:
            C_sweep = C_PML
        layer_name = "layer"+str(layer-1)
        definedon_PML_move = "layer"+str(layer-1)
        rad_min = r_layer[-1-layer]
        rad_max = r_layer[-1-layer+1]
        
        eta = rad_max-rad_min
        movepml_sigma = IfPos(y-rad_min,IfPos(rad_max-y,C_sweep/eta*(-rad_max+y+eta)**2/eta**2,0),0)

        movepml_trafo = CoefficientFunction( (x,
                                             y+(1j/f0)*(C_sweep/3)*(eta*movepml_sigma/C_sweep)**(3/2) ) )
        movepml_jac = CoefficientFunction((1.0,
                                           0.0,
                                           0.0,
                                           1+(1j/f0)*movepml_sigma 
                                           ),
                                      dims=(2,2))   
        
        pml_move = pml.Custom(trafo=movepml_trafo, jac = movepml_jac ) 
        
        ba_dofs = BitArrayFromList(list( set(layer2dof[layer-1]) - set(IF2dof[layer])  ),mesh,fes,elements=False)
        ba_els = BitArrayFromList(layer2el[layer-1],mesh,fes,elements=True) 
       
        if layer > 1:
            mesh.SetPML(pml_move,1)
        else:
            if use_PML:
                mesh.SetPML(pml_move,1)
        a_move = BilinearForm (fes, symmetric=False, check_unused=False)
        a_move += SymbolicBFI ( ( y**2*(rho*c**2)*grad(u)[1]*grad(v)[1] +  (rho*c**2)*grad(u)[0]*grad(v)[0])*weight_factor ,definedonelements= ba_els,bonus_intorder =  bonus_intorder)  
        a_move += SymbolicBFI ( -rho*omega**2*y**2*u*v*weight_factor ,definedonelements= ba_els,bonus_intorder =  bonus_intorder)  
        a_move.Assemble()
        layermat_moving_PML[layer] = a_move.mat  
        layerinverse_moving_PML[layer] = a_move.mat.Inverse(ba_dofs,inverse=solver)
        if layer > 1: 
            mesh.UnSetPML(1) 
        else:
            if use_PML:
                mesh.SetPML(pml_move,1)

#########################################################################
# Transmission operators of DOSM
#########################################################################

tmp_DtN = gfu.vec.CreateVector()
tmp1_DtN = gfu.vec.CreateVector()

DtN_eigenbasis = [ [] for i in range(n_layers)]
M_IF_matrices = [ [] for i in range(n_layers)]

for layer_nr in range(1,n_layers):
    if_idx = layer_nr
    n_len = len(IF2dof[if_idx])  
    
    DtN_in_discrete_eigenbasis = np.zeros((len(lam_layer[layer_nr]),(len(lam_layer[layer_nr]))),dtype=complex)
    
    ba_dofs = BitArrayFromList(IF2dof[if_idx],mesh,fes,elements=False)
    m = BilinearForm (fes, symmetric=True, check_unused=False)
    m += SymbolicBFI ( u*v*weight_theta, definedon=mesh.Boundaries("layer"+str(if_idx)+"-bnd") )
    m.Assemble()
    M_IF_matrices[layer_nr] = m.mat

    for nr,eval_nr in enumerate(lam_layer[layer_nr]):   
        DtN_in_discrete_eigenbasis[nr,nr] = DtN_coeff_layer[layer_nr][nr]
    DtN_eigenbasis[layer_nr] = DtN_in_discrete_eigenbasis.copy()


def DtN_semi_discrete(layer,v_in,v_out):
    proj_IF = Projector(BitArrayFromList(IF2dof[layer],mesh,fes,elements=False),False)
    tmp_DtN.data = v_in
    tmp_DtN.data -= proj_IF * tmp_DtN

    data_in_eigenbasis = np.linalg.solve(lam_vec_layer[layer], tmp_DtN.FV().NumPy()[IF2dof[layer]] )
    tmp_DtN.FV().NumPy()[IF2dof[layer]] = data_in_eigenbasis[:]

    result_in_eigenbasis = np.diag(np.diag(DtN_eigenbasis[layer])) @ tmp_DtN.FV().NumPy()[IF2dof[layer]]
    back_trafo = lam_vec_layer[layer] @ result_in_eigenbasis

    v_out.FV().NumPy()[IF2dof[layer]] = back_trafo[:]
    tmp_DtN.data =  M_IF_matrices[layer] * v_out
    v_out.data = tmp_DtN
    v_out.data -= proj_IF * v_out

if use_moving_PML:
    def DtN_moving_PML(layer,v_in,v_out):
       
        proj_IF = Projector(BitArrayFromList(IF2dof[layer],mesh,fes,elements=False),False)
        proj_IF_outer = Projector(BitArrayFromList(IF2dof[layer],mesh,fes,elements=False),False)
        proj_outer = Projector(BitArrayFromList(layer2all_outer_dof[layer],mesh,fes,elements=False),False)

        tmp_DtN.data = v_in

        tmp_DtN.data -= proj_IF * tmp_DtN
        tmp1_DtN.data = layermat_moving_PML[layer] * tmp_DtN 
        tmp1_DtN.data -= proj_IF * tmp1_DtN
        v_out.data  = tmp1_DtN

        tmp1_DtN.data =  layermat_moving_PML[layer] * tmp_DtN
        tmp1_DtN.data -= proj_outer * tmp1_DtN
        tmp_DtN.data =  layerinverse_moving_PML[layer] * tmp1_DtN

        tmp_DtN.data -= proj_outer * tmp_DtN

        rhs_tmp = gfu.vec.CreateVector()
        tmp1_DtN.data = layermat_moving_PML[layer] * tmp_DtN

        tmp1_DtN.data -= proj_IF * tmp1_DtN
        v_out.data -= tmp1_DtN

    def DtN_from_moving_PML(layer,v_in,v_out):
        proj_IF_outer = Projector(BitArrayFromList(IF2dof[layer],mesh,fes,elements=False),False)
        DtN_moving_PML(layer,v_in,v_out)
        v_out.data -= proj_IF_outer * v_out

#########################################################################
# Transmission operators of DOSM
#########################################################################

def Q_outer(layer,v_in,v_out):
    v_out.data = v_in

def P_outer(layer,v_in,v_out): 
    if use_moving_PML:
        DtN_moving_PML(layer,v_in,v_out)
    else: 
        DtN_semi_discrete(layer,v_in,v_out)

#########################################################################
# Assembling the matrices for solving the system on layer interfaces
#########################################################################

IF_lus = [None for i in range(n_layers)] 
for layer in range(1,n_layers):
    atmp = gfu.vec.CreateVector()
    atmp1 = gfu.vec.CreateVector()
    atmp_in = gfu.vec.CreateVector()
    atmp_out = gfu.vec.CreateVector()

    proj_IF_outer = Projector(BitArrayFromList(IF2dof[layer],mesh,fes,elements=False),False)
    proj_inner = Projector(BitArrayFromList(Interior2dof[layer],mesh,fes,elements=False),False)

    def apply_matvec(v):
        atmp_in.FV().NumPy()[:] = 0.0
        if len(v.shape) > 1:
            atmp_in.FV().NumPy()[IF2dof[layer]] = v[:,0]
        else:
            atmp_in.FV().NumPy()[IF2dof[layer]] = v

        P_outer(layer,atmp_in,atmp_out)

        atmp.data = atmp_in
        atmp.data -= proj_IF_outer * atmp
        atmp1.data = layermat[layer] * atmp
        atmp.data = atmp1
        atmp.data -= proj_IF_outer * atmp
        Q_outer(layer,atmp,atmp1)
        atmp_out.data += atmp1
            
        atmp.data = atmp_in
        atmp.data -= proj_IF_outer * atmp
        atmp1.data = layermat[layer] * atmp
        atmp1.data -= proj_inner * atmp1
        atmp.data = layerinverse[layer] * atmp1
        atmp.data -= proj_inner * atmp
        atmp1.data = layermat[layer] * atmp
        atmp1.data -= proj_IF_outer * atmp1
        Q_outer(layer,atmp1,atmp)
        atmp_out.data -= atmp
             
        return atmp_out.FV().NumPy()[IF2dof[layer]]
            
    n_len = len(IF2dof[layer])
    layer_IF_mat = np.zeros((n_len,n_len),dtype=complex)

    ei = np.zeros(n_len,dtype = complex) 
    for ii in range(n_len):
        ei[:] = 0.0
        ei[ii] = 1.0
        rei = apply_matvec(ei)
        layer_IF_mat[:,ii] = rei[:]
    
    IF_lus[layer] = lu_factor(layer_IF_mat)


#########################################################################
# DOSM algorithm in matrix form 
# see Algorithm 2 in the paper 
# "A CLASS OF ITERATIVE SOLVERS FOR THE HELMHOLTZ EQUATION:
#  FACTORIZATIONS. SWEEPING PRECONDITIONERS, SOURCE TRANSFER,
#  SINGLE LAYER POTENTIALS, POLARIZED TRACES, AND OPTIMIZED
#  SCHWARZ METHODS (M.J. Gander and H. Zhang) 
#########################################################################

def apply_DOSM_mats(layer,v_in,v_out):
    # v_in:   input vector
    # v_out:  output vector

    # clear output vector
    v_out.FV().NumPy()[:] = 0.0 

    rhs_IF_outer = gfu.vec.CreateVector()
    rhs_inner = gfu.vec.CreateVector()
    tmp = gfu.vec.CreateVector()
    tmp1 = gfu.vec.CreateVector()
    tmp_in = gfu.vec.CreateVector()
    tmp_out = gfu.vec.CreateVector()

    proj_IF_outer = Projector(BitArrayFromList(IF2dof[layer],mesh,fes,elements=False),False)
    proj_inner = Projector(BitArrayFromList(Interior2dof[layer],mesh,fes,elements=False),False)
 
    # Dirichlet b.c. on inner interface
    if layer + 1 < n_layers:
        proj_IF_inner = Projector(BitArrayFromList(IF2dof[layer+1],mesh,fes,elements=False),False)
        v_out.data = v_in
        v_out.data -= proj_IF_inner * v_out.data
   
    if layer == 0:
        # solve directly on layer 0
        inner_dof_plus_outer_IF = list( set(layer2dof[0]) - set(IF2dof[1])  )
        proj_inner_dof_plus_outer_IF = Projector(BitArrayFromList(inner_dof_plus_outer_IF,mesh,fes,elements=False),False)
        proj_IF_inner = Projector(BitArrayFromList(IF2dof[layer+1],mesh,fes,elements=False),False)
        rhs_inner.data = v_in
        rhs_inner.data -=  proj_inner_dof_plus_outer_IF * v_in
        tmp.data = v_out.data 
        tmp.data -= proj_IF_inner * v_out.data
        tmp1.data = layermat[layer] * tmp
        tmp1.data -= proj_inner_dof_plus_outer_IF * tmp1
        rhs_inner.data -= tmp1
        tmp.data = layerinverse[layer] * rhs_inner
        tmp.data -= proj_inner_dof_plus_outer_IF * tmp
        v_out.data += tmp

    else:
        # solve iteratively for unknowns on outer interface 
        # 1. set up the right hand side
        rhs_IF_outer.data = v_in
        rhs_IF_outer.data -= proj_inner * rhs_IF_outer  
        if  layer + 1 < n_layers:
            tmp.data = layermat[layer] * v_out
            tmp.data -= proj_inner * tmp
            rhs_IF_outer.data -= tmp
        tmp.data = layerinverse[layer] * rhs_IF_outer
        rhs_IF_outer.data = tmp
        rhs_IF_outer.data -= proj_inner * rhs_IF_outer
        tmp.data = layermat[layer] * rhs_IF_outer 
        rhs_IF_outer.data = tmp
        rhs_IF_outer.data -= proj_IF_outer * rhs_IF_outer
        Q_outer(layer,rhs_IF_outer,tmp)
        rhs_IF_outer.data = v_in
        rhs_IF_outer.data -= proj_IF_outer * rhs_IF_outer
        rhs_IF_outer.data -= tmp

        # 2. set up left hand side as linear operator
        def apply_matvec(v):
            tmp_in.FV().NumPy()[:] = 0.0
            tmp_in.FV().NumPy()[IF2dof[layer]] = v

            P_outer(layer,tmp_in,tmp_out)
            tmp.data = tmp_in
            tmp.data -= proj_IF_outer * tmp
            tmp1.data = layermat[layer] * tmp
            tmp.data = tmp1
            tmp.data -= proj_IF_outer * tmp
            Q_outer(layer,tmp,tmp1)
            tmp_out.data += tmp1
            
            tmp.data = tmp_in
            tmp.data -= proj_IF_outer * tmp
            tmp1.data = layermat[layer] * tmp
            tmp1.data -= proj_inner * tmp1
            tmp.data = layerinverse[layer] * tmp1
            tmp.data -= proj_inner * tmp
            tmp1.data = layermat[layer] * tmp
            tmp1.data -= proj_IF_outer * tmp1
            Q_outer(layer,tmp1,tmp)
            tmp_out.data -= tmp

            return tmp_out.FV().NumPy()[IF2dof[layer]]

 
        f_gmres = np.zeros(len(IF2dof[layer]),dtype = complex)
        f_gmres[:] = rhs_IF_outer.FV().NumPy()[IF2dof[layer]]

        tmp.FV().NumPy()[IF2dof[layer]] = lu_solve(IF_lus[layer],f_gmres)[:]

                                                    
        tmp.data -= proj_IF_outer * tmp
        v_out.data += tmp

           
        # solve for inner unknowns on layer
        # 1. set up the right hand side
        rhs_inner.data =  v_in
        rhs_inner.data -= proj_inner * rhs_inner
        if layer < n_layers -1:
            proj_IF_inner = Projector(BitArrayFromList(IF2dof[layer+1],mesh,fes,elements=False),False)
            tmp.data = v_in
            tmp.data -= proj_IF_inner * v_in
            tmp1.data = layermat[layer] * tmp
            tmp1.data -= proj_inner * tmp1
            rhs_inner.data -= tmp1
        tmp.data = v_out
        tmp.data -= proj_IF_outer * tmp
        tmp1.data = layermat[layer] * tmp
        
        tmp1.data -=  proj_inner * tmp1
        rhs_inner.data -= tmp1
        
        # 2. Solve system for inner dofs
        tmp.data = layerinverse[layer] * rhs_inner
        tmp.data -= proj_inner * tmp
        v_out.data += tmp

def ApplyDOSM(res,u_new,u_old):        
    unew = u_new.CreateVector()
    uold = u_old.CreateVector()
    uresult = u_old.CreateVector()
    unew[:] = 0.0
    uold.data = u_old
        
    ufwd = u_new.CreateVector()
    ufwd[:] = 0.0
        
    rhs = res.CreateVector()
    tmp = res.CreateVector()
    tmp1 = res.CreateVector()
    rhs.data = res
    rhs_layer = res.CreateVector()
        
    # forward sweep
    rhs_fwd = [res.CreateVector() for i in range(n_layers)]
    for l in range(n_layers):

        if l == 0:
            inner_dof_plus_outer_IF = list( set(layer2dof[0]) - set(IF2dof[1])  )
            proj_inner_dof_plus_outer_IF = Projector(BitArrayFromList(inner_dof_plus_outer_IF,mesh,fes,elements=False),False)
            rhs_layer.data = res
            rhs_layer.data -= proj_inner_dof_plus_outer_IF * rhs_layer
        else:
            proj_inner = Projector(BitArrayFromList(Interior2dof[l],mesh,fes,elements=False),False)
            rhs_layer.data = res
            rhs_layer.data -= proj_inner * rhs_layer
          
        if l < n_layers-1:
            proj_IF_inner = Projector(BitArrayFromList(IF2dof[l+1],mesh,fes,elements=False),False)
            tmp.data = uold
            tmp.data -= proj_IF_inner * tmp
            rhs_layer.data += tmp

        if l > 0:
            proj_IF_outer = Projector(BitArrayFromList(IF2dof[l],mesh,fes,elements=False),False)
            proj_inner_prev = Projector(BitArrayFromList(Interior2dof[l-1],mesh,fes,elements=False),False)
            tmp.data = unew.data 
            tmp.data -= proj_IF_outer * tmp
            tmp1.data = layermat_outer[l] * tmp
            tmp1.data -= proj_IF_outer * tmp1
            Q_outer(l,tmp1,tmp)
            rhs_layer.data -= tmp
            tmp.data = unew.data 
            tmp.data -= proj_IF_outer * tmp
            P_outer(l,tmp,tmp1)
            rhs_layer.data += tmp1
            tmp.data = unew.data 
            tmp.data -= proj_inner_prev * tmp
            tmp1.data = layermat[l-1] * tmp
            tmp1.data -= proj_IF_outer * tmp1
            tmp.data  = res
            tmp.data -= proj_IF_outer * tmp
            tmp.data -= tmp1
            Q_outer(l,tmp,tmp1)
            rhs_layer.data += tmp1

            
        rhs_fwd[l].data = rhs_layer
        if l < n_layers-1:  
            apply_DOSM_mats(l,rhs_layer,uresult)  
            unew.FV().NumPy()[layer2dof[l]] = uresult.FV().NumPy()[layer2dof[l]]
            gfu_DOSM_interm.vec.data = unew


    ufwd.data = unew
         
    # backward sweep 
    for l in range(n_layers-1,-1,-1):
        rhs_layer.data = rhs_fwd[l]
        if l != n_layers-1:   
            rhs_layer.FV().NumPy()[IF2dof[l+1]] = unew.FV().NumPy()[IF2dof[l+1]]
        apply_DOSM_mats(l,rhs_layer,uresult)  
        unew.FV().NumPy()[layer2dof[l]] = uresult.FV().NumPy()[layer2dof[l]]
        gfu_DOSM_interm.vec.data = unew
            
    u_new.data = unew

#########################################################################
# Class definition of the preconditioner
#########################################################################

gfu_DOSM_in.vec[:] = 0.0
gfu_DOSM_out.vec[:] = 0.0
DOSM_error = [] 

class DOSMPrecond(BaseMatrix):
    
    def IsComplex(self):
        return True

    def Height(self):
        return len(gfu.vec)

    def Width(self):
        return len(gfu.vec)
    
    def Mult(self, x, y):
        gfu_DOSM_in.vec[:] = 0.0
        ApplyDOSM(x,y,gfu_DOSM_in.vec)
    
C =  DOSMPrecond()


#########################################################################
# One application of preconditioner
#########################################################################

ApplyDOSM(f.vec,gfu_DOSM_out.vec,gfu_DOSM_in.vec)
l2_norm = sqrt (Integrate ( (udirect.real)*(udirect.real) + (udirect.imag)*(udirect.imag), mesh))
l2_error = sqrt (Integrate ( (gfu_DOSM_out.real-udirect.real)*(gfu_DOSM_out.real-udirect.real) + (gfu_DOSM_out.imag-udirect.imag)*(gfu_DOSM_out.imag-udirect.imag), mesh))
print('Relative L2-error after one application of preconditioner: %.2e' % Decimal(l2_error/l2_norm))

#########################################################################
# Preconditioned GMRES iteration
#########################################################################

res = []
global cnt2
cnt2 = 0

def callback_full(v):
    global cnt2
    gfu_GMRES.vec.data = v
    l2_error_iter = sqrt (Integrate ( (gfu_GMRES.real-udirect.real)*(gfu_GMRES.real-udirect.real) + (gfu_GMRES.imag-udirect.imag)*(gfu_GMRES.imag-udirect.imag), mesh))
    res.append(l2_error_iter)
    #print("Iter = {0}, L2-error = {1}".format(cnt2,l2_error_iter))
    cnt2 += 1

print("Start GMRES iteration with DOSM preconditioner")
gfu_GMRES.vec.data = GMRes(a.mat, f.vec,freedofs=fes.FreeDofs(), pre=C, maxsteps = 1000, tol = tol,
          callback=callback_full, restart=None, startiteration=0, printrates=True)

print("L2-errors in iterations = {0}".format(res))
l2_error = sqrt (Integrate ( (gfu_GMRES.real-udirect.real)*(gfu_GMRES.real-udirect.real) + (gfu_GMRES.imag-udirect.imag)*(gfu_GMRES.imag-udirect.imag), mesh))
print ("Final L2-error :", l2_error )

